/** Copyright 2017 Amazon.com. All rights reserved. */

#pragma once

#include <cstddef>

#include <algorithm>
#include <condition_variable>
#include <utility>
#include <curl/curl.h>
#include <string>
#include <mutex>
#include <fstream>

#include "com/amazonaws/kinesis/video/client/Include.h"
#include "CallbackProvider.h"
#include "ResponseAcceptor.h"

namespace com { namespace amazonaws { namespace kinesis { namespace video {

/**
* Internal class to wrap a condition variable that signals the curl callback that data is ready and to try to read
* data off Kinesis Video PIC's internal data buffer in order to post it over the network.
*
* Additionally, this class contains all the state that is needed by the postBodyStreamingFunc and is passed in as
* the void* custom_data parameter to that class.
*/
class OngoingStreamState : public ResponseAcceptor {
public:
    OngoingStreamState(CallbackProvider* callback_provider,
                       UPLOAD_HANDLE upload_handle,
                       STREAM_HANDLE stream_handle,
                       std::string stream_name,
                       bool debug_dump_file = false);

    ~OngoingStreamState() = default;

    /**
     *
     * Notify the condition variable that data was received and is ready for transit over the network thread.
     *
     * @param duration_available duration of the data available in 100ns.
     * @param size_available the size available for reading off Kinesis Video PIC's buffer in bytes.
     */
    void noteDataAvailable(UINT64 duration_available, UINT64 size_available);

    /**
     * Returns the current stream handle we are streaming for
     *
     * @return STREAM_HANDLE
     */
    STREAM_HANDLE getStreamHandle() const {
        return stream_handle_;
    }

    /**
     * Returns the callback provider
     *
     * @return shared pointer to CallbackProvider object
     */
    CallbackProvider* getCallbackProvider() const {
        return callback_provider_;
    }

    /**
     * Returns the stream name
     * @return stream name
     */
    std::string getStreamName() {
        return stream_name_;
    }

    /**
     * Returns whether we have reached end-of-stream
     */
    bool isEndOfStream() {
        return end_of_stream_;
    }

    /**
     * Signals the end of the stream
     */
    void endOfStream() {
        end_of_stream_ = true;
    }

    /**
     * Returns whether CURL is in shutdown state
     */
    bool isShutdown() {
        return shutdown_;
    }

    /**
     * Signals the CURL shutdown
     */
    void shutdown() {
        end_of_stream_ = true;
        shutdown_ = true;
    }

    /**
     * Returns whether CURL has been terminated
     */
    bool isTerminated() {
        return terminated_;
    }

    /**
     * Close debug dump file and set termination flag.
     */
    void terminate() {
        if (debug_dump_file_ && debug_dump_file_stream_.is_open()) {
            debug_dump_file_stream_.close();
        }

        terminated_ = true;
    }

    /**
     * Sets the current CURL response object
     */
    void setResponse(STREAM_HANDLE stream_handle, std::shared_ptr<Response> response) override {
        curl_response_ = response;
    }

    /**
     * Gets the current CURL response object
     */
    std::shared_ptr<Response> getResponse() {
        return curl_response_;
    }

    /**
     * Returns the stream upload handle
     */
    UPLOAD_HANDLE getUploadHandle() {
        return upload_handle_;
    }

    /**
     * Pulse another upload handles
     */
    void pulseUploadHandle(uint64_t upload_handle);

    /**
     * Unpause the upload
     */
    bool unPause();

    /**
     * An implementation of the CURLOPT_READFUNCTION callback: https://curl.haxx.se/libcurl/c/CURLOPT_READFUNCTION.html.
     *
     * From the documentation:
     * This callback function gets called by libcurl as soon as it needs to read data in order to send it to the peer -
     * like if you ask it to upload or post data to the server.
     * The data area pointed at by the pointer buffer should be filled up with at most item_size multiplied with n_items
     * number
     * of bytes by your function.
     * This function must then return the actual number of bytes that it stored in that memory area. Returning 0 will
     * signal end-of-file to the library and cause it to stop the current transfer.
     *
     * @param buffer Curl buffer to be filled
     * @param item_size size of each item in the buffer in bytes
     * @param n_items Max number of items of size item_size that can fit in the buffer
     * @return The number of bytes written to the buffer
     */
    size_t postBodyStreamingReadFunc(char *buffer, size_t item_size, size_t n_items);

    /**
     * An implementation of the CURLOPT_HEADERFUNCTION callback:
     * https://curl.haxx.se/libcurl/c/CURLOPT_HEADERFUNCTION.html.
     *
     * From the documentation:
     * This function gets called by libcurl as soon as it has received header data. The header callback will be called
     * once for each header and only complete header lines are passed on to the callback. Parsing headers is very easy
     * using this. The size of the data pointed to by buffer is size multiplied with nmemb. Do not assume that the
     * header line is zero terminated! The pointer named userdata is the one you set with the CURLOPT_HEADERDATA
     * option. This callback function must return the number of bytes actually taken care of. If that amount differs
     * from the amount passed in to your function, it'll signal an error to the library. This will cause the transfer
     * to get aborted and the libcurl function in progress will return CURLE_WRITE_ERROR.
     *
     * @param buffer Curl buffer to be filled
     * @param item_size size of each item in the buffer in bytes
     * @param n_items Max number of items of size item_size that can fit in the buffer
     * @return The number of bytes written to the buffer
     */
    size_t postHeaderReadFunc(char *buffer, size_t item_size, size_t n_items);

    /**
     * An implementation of the CURLOPT_READFUNCTION callback: https://curl.haxx.se/libcurl/c/CURLOPT_WRITEFUNCTION.html.
     *
     * From the documentation:
     * This callback function gets called by libcurl as soon as there is data received that needs to be saved. ptr
     * points to the delivered data, and the size of that data is size multiplied with nmemb.
     *
     * The callback function will be passed as much data as possible in all invokes, but you must not make any assumptions.
     * It may be one byte, it may be thousands. The maximum amount of body data that will be passed to the write
     * callback is defined in the curl.h header file: CURL_MAX_WRITE_SIZE (the usual default is 16K).
     *
     * If CURLOPT_HEADER is enabled, which makes header data get passed to the write callback, you can get up to
     * CURL_MAX_HTTP_HEADER bytes of header data passed into it. This usually means 100K.
     *
     * @param buffer Curl buffer to read from
     * @param item_size size of each item in the buffer in bytes
     * @param n_items Max number of items of size item_size available in the buffer
     * @return The number of bytes read from the buffer
     */
    size_t postBodyStreamingWriteFunc(char *buffer, size_t item_size, size_t n_items);

private:

    /**
     * Stream handle
     */
    STREAM_HANDLE stream_handle_;

    /**
     * The main owner object
     */
    CallbackProvider* callback_provider_;

    /**
     * Stream name
     */
    std::string stream_name_;

    /**
     * Stream upload handle
     */
    UPLOAD_HANDLE upload_handle_;

    /**
     * Whether we have reached end-of-stream and the connection needs to be closed
     */
    volatile bool end_of_stream_;

    /**
     * CURL is shutting down
     */
    volatile bool shutdown_;

    /**
     * Ongoing CURL response object
     */
    std::shared_ptr<Response> curl_response_;

    /**
     * Debug output to a file
     */
    std::ofstream debug_dump_file_stream_;

    /**
     * Whether the debug dump file is opened and needs to be closed.
     */
    bool debug_dump_file_;

    /**
     * Whether data has become available to send to Kinesis Video PIC.
     */
    volatile bool data_available_;

    /**
     * Whether the session is waiting for the last persisted ack.
     */
    volatile bool awaiting_persisted_ack_;

    /**
     * Whether zero has been returned in the curl callback to terminate the underlying curl object.
     */
    volatile bool terminated_;
};

} // namespace video
} // namespace kinesis
} // namespace amazonaws
} // namespace com

